@@ -911,7 +911,7 @@ class CreateIndexOp(MigrateOperation):
         operations: Operations,
         index_name: Optional[str],
         table_name: str,
-        columns: Sequence[Union[str, TextClause, Function]],
+        columns: Sequence[Union[str, TextClause, Function[Any]]],
         schema: Optional[str] = None,
         unique: bool = False,
         **kw: Any,
@@ -1885,7 +1885,7 @@ class AlterColumnOp(AlterTableOp):
         column_name: str,
         nullable: Optional[bool] = None,
         comment: Union[str, Literal[False]] = False,
-        server_default: Union[Function, bool] = False,
+        server_default: Union[Function[Any], bool] = False,
         new_column_name: Optional[str] = None,
         type_: Optional[Union[TypeEngine, Type[TypeEngine]]] = None,
         existing_type: Optional[Union[TypeEngine, Type[TypeEngine]]] = None,
